import typing

import sys


from pygments.token import *
from pygments.lexer import Lexer
import antlr4
from .KeYLexer import KeYLexer as AKL

# antlr4 -Dlanguage=Python3 MyGrammar.g4


SPECIAL = {
    AKL.WS: Whitespace,
    AKL.INT_LITERAL: Number,
    AKL.STRING_LITERAL: String,
    AKL.QUOTED_STRING_LITERAL: String.Double,
    AKL.DOC_COMMENT: String.Doc,
    AKL.FLOAT_LITERAL: Number.Float,
    AKL.CHAR_LITERAL: String.Char,
    AKL.COMMENT_END: Comment.Multiline,
    AKL.ML_COMMENT: Comment.Multiline,
    AKL.SL_COMMENT: Comment.Single,
    AKL.IDENT: Name.Variable,
    AKL.FALSE: Literal.Bool,
    AKL.TRUE: Literal.Bool,
    AKL.DOUBLE_LITERAL: Literal.Float,
    AKL.REAL_LITERAL: Literal.Float,
    AKL.ERROR_CHAR: Error,
    AKL.BIN_LITERAL: Number.Bin,
    AKL.HEX_LITERAL: Number.Hex,
}

OPERATORS = frozenset((AKL.LESS, AKL.LESSEQUAL, AKL.LGUILLEMETS, AKL.EQV, AKL.PRIMES,
                      AKL.EXP, AKL.TILDE, AKL.PERCENT, AKL.STAR, AKL.MINUS, AKL.PLUS, AKL.GREATER, AKL.GREATEREQUAL, AKL.SLASH,  AKL.ASSIGN,
                      AKL.AT, AKL.PARALLEL, AKL.OR, AKL.AND, AKL.NOT, AKL.IMP, AKL.EQUALS, AKL.NOT_EQUALS, AKL.SEQARROW, AKL.DOTRANGE))

PUNCTATION = frozenset((AKL.COMMA, AKL.LPAREN, AKL.RPAREN, AKL.LBRACE, AKL.RBRACE, AKL.LBRACKET, AKL.RBRACKET, AKL.SEMI,
                       AKL.COLON, AKL.DOUBLECOLON,
                       AKL.EMPTYBRACKETS, AKL.RGUILLEMETS, AKL.DOT))

KEYWORDS = frozenset((AKL.MODALITY, AKL.SORTS,
                     AKL.GENERIC, AKL.PROXY, AKL.EXTENDS, AKL.ONEOF, AKL.ABSTRACT, AKL.SCHEMAVARIABLES, AKL.SCHEMAVAR, AKL.MODALOPERATOR, AKL.PROGRAM, AKL.FORMULA,
                     AKL.TERM, AKL.UPDATE, AKL.VARIABLES, AKL.VARIABLE, AKL.SKOLEMTERM, AKL.SKOLEMFORMULA, AKL.TERMLABEL, AKL.MODIFIES, AKL.PROGRAMVARIABLES,
                     AKL.STORE_TERM_IN, AKL.STORE_STMT_IN, AKL.HAS_INVARIANT, AKL.GET_INVARIANT, AKL.GET_FREE_INVARIANT, AKL.GET_VARIANT,
                     AKL.IS_LABELED, AKL.SAME_OBSERVER, AKL.VARCOND, AKL.APPLY_UPDATE_ON_RIGID, AKL.DEPENDINGON, AKL.DISJOINTMODULONULL,
                     AKL.DROP_EFFECTLESS_ELEMENTARIES, AKL.DROP_EFFECTLESS_STORES, AKL.SIMPLIFY_IF_THEN_ELSE_UPDATE, AKL.ENUM_CONST, AKL.FREELABELIN,
                     AKL.HASSORT, AKL.FIELDTYPE, AKL.FINAL, AKL.ELEMSORT, AKL.HASLABEL, AKL.HASSUBFORMULAS, AKL.ISARRAY, AKL.ISARRAYLENGTH, AKL.ISCONSTANT,
                     AKL.ISENUMTYPE, AKL.ISINDUCTVAR, AKL.ISLOCALVARIABLE, AKL.ISOBSERVER, AKL.DIFFERENT, AKL.METADISJOINT, AKL.ISTHISREFERENCE,
                     AKL.DIFFERENTFIELDS, AKL.ISREFERENCE, AKL.ISREFERENCEARRAY, AKL.ISSTATICFIELD, AKL.ISINSTRICTFP, AKL.ISSUBTYPE, AKL.EQUAL_UNIQUE,
                     AKL.NEW, AKL.NEW_TYPE_OF, AKL.NEW_DEPENDING_ON, AKL.HAS_ELEMENTARY_SORT, AKL.NEWLABEL, AKL.CONTAINS_ASSIGNMENT, AKL.NOT_,
                     AKL.NOTFREEIN, AKL.SAME, AKL.STATIC, AKL.STATICMETHODREFERENCE, AKL.MAXEXPANDMETHOD, AKL.STRICT, AKL.TYPEOF, AKL.INSTANTIATE_GENERIC,
                     AKL.FORALL, AKL.EXISTS, AKL.SUBST, AKL.IF, AKL.IFEX, AKL.THEN, AKL.ELSE, AKL.INCLUDE, AKL.INCLUDELDTS, AKL.CLASSPATH, AKL.BOOTCLASSPATH,
                     AKL.NODEFAULTCLASSES, AKL.JAVASOURCE, AKL.WITHOPTIONS, AKL.OPTIONSDECL, AKL.KEYSETTINGS, AKL.PROFILE,
                     AKL.SAMEUPDATELEVEL, AKL.INSEQUENTSTATE, AKL.ANTECEDENTPOLARITY, AKL.SUCCEDENTPOLARITY, AKL.CLOSEGOAL,
                     AKL.HEURISTICSDECL, AKL.NONINTERACTIVE, AKL.DISPLAYNAME, AKL.HELPTEXT, AKL.REPLACEWITH, AKL.ADDRULES,
                     AKL.ADDPROGVARS, AKL.HEURISTICS, AKL.FIND, AKL.ADD, AKL.ASSUMES, AKL.TRIGGER, AKL.AVOID, AKL.PREDICATES,
                     AKL.FUNCTIONS, AKL.TRANSFORMERS, AKL.UNIQUE, AKL.RULES, AKL.AXIOMS, AKL.PROBLEM, AKL.CHOOSECONTRACT,
                     AKL.PROOFOBLIGATION, AKL.PROOF, AKL.PROOFSCRIPT, AKL.CONTRACTS, AKL.INVARIANTS, AKL.LEMMA, AKL.IN_TYPE,
                     AKL.IS_ABSTRACT_OR_INTERFACE, AKL.CONTAINERTYPE, AKL.UTF_PRECEDES, AKL.UTF_IN, AKL.UTF_EMPTY, AKL.UTF_UNION,
                     AKL.UTF_INTERSECT, AKL.UTF_SUBSET, AKL.UTF_SETMINUS,
                     AKL.MODALITYD, AKL.MODALITYB, AKL.MODALITYBB, AKL.MODAILITYGENERIC1,
                     AKL.MODAILITYGENERIC2, AKL.MODAILITYGENERIC3, AKL.MODAILITYGENERIC4, AKL.MODAILITYGENERIC5, AKL.MODAILITYGENERIC6, AKL.MODAILITYGENERIC7,
                     AKL.MODALITYD_END, AKL.MODALITYD_STRING, AKL.MODALITYD_CHAR, AKL.MODALITYG_END, AKL.MODALITYB_END, AKL.MODALITYBB_END))


__all__ = ('KeYLexer')


class KeYLexer(Lexer):
    name = 'KeY'
    aliases = ['key']
    filenames = ['*.key']

    def __init__(self, **options) -> None:
        super().__init__(**options)
        self.table = [None] * (AKL.MODALITYBB_END+1)
        for typ, style in SPECIAL.items():
            self.table[typ] = style

        styles = (Operator, Punctuation, Keyword)
        for style, typseq in zip(styles, (OPERATORS, PUNCTATION, KEYWORDS)):
            for typ in typseq:
                self.table[typ] = style

    def get_tokens_unprocessed(self, text: str):
        input_stream = antlr4.InputStream(text)
        lexer = AKL(input_stream)
        tokens: typing.List[antlr4.Token] = lexer.getAllTokens()

        for idx, token in enumerate(tokens):
            if token.type == AKL.INT_LITERAL:
                for lookahead in tokens[idx+1:]:
                    if lookahead.channel == 2:
                        continue
                    if lookahead.type == AKL.LPAREN:
                        token.setType(AKL.IDENT)
                    break

            clazz = self.class_for_type(token.type)
            if clazz is None:
                continue

            off = token.start
            value = token.text
            yield (off, clazz, value)

    def class_for_type(self, token: int):
        if v := self.table[token]:
            return v
        else:
            return Error




from pygments.lexer import Lexer
from pygments.lexers.jvm import JavaLexer

#import pygments.token as T
from pygments.token import *

from .JML import JML

from typing import List, Dict

import antlr4

#Comment = T.Comment.Jml
#Whitespace = T.Whitespace.Jml
#Error = T.Error.Jml
#Punctuation = T.Punctuation.Jml
#Keyword = T.Keyword.Jml
#Name = T.Name.Jml
#Operator =T.Operator.Jml
#Number = T.Number.Jml


class JMLLexer(Lexer):
    name = 'JML'
    aliases = ['jml', 'java-modeling-language']
    filenames = ['*.java', '*.jml']

    def __init__(self, **options) -> None:
        active_keys = options.get("active-keys", list())
        self.conditional_checker = JmlConditionalChecker(active_keys)
        super().__init__(**options)
        self.java_lex = JavaLexer(**options)

    def get_tokens_unprocessed(self, text: str):
        for (index, kind, value) in self.java_lex.get_tokens_unprocessed(text):
            if kind == Comment.Single or kind == Comment.Multiline:
                yield from self.expand_jml(index, kind, value)
            else:
                yield (index, kind, value)

    def expand_jml(self, index, kind, text):
        # analyze the document and return semantic tokens
        # if (token.type == JML.JML_SET_KEY):
        #    conditionalChecker.setKeysFromToken(token)
        #    continue

        if not self.conditional_checker.is_active(text):
            yield (index, kind, text)
        else:
            start = text.find("@") + 1
            offset = index + start

            yield (index, Comment.Preproc, text[:start])

            lexer = JML(antlr4.InputStream(text[start:]))
            lexer._mode = JML.jmlMode

            parenthesisLevel = 0
            bracesLevel = 0
            bracketLevel = 0
            waitForToplevelSemicolon = False

            for token in lexer.getAllTokens():
                offset = token.start + start
                level0 = bracketLevel == 0 and bracesLevel == 0 and parenthesisLevel == 0
                toplevel = not waitForToplevelSemicolon and level0

                match token.type:
                    case JML.LPAREN: parenthesisLevel += 1
                    case JML.RPAREN: parenthesisLevel -= 1
                    case JML.LBRACE: bracesLevel += 1
                    case JML.RBRACE: bracesLevel -= 1
                    case JML.LBRACE: bracketLevel += 1
                    case JML.RBRACE: bracketLevel -= 1
                    case JML.SEMI:
                        if level0:
                            waitForToplevelSemicolon = False

                if token.type in TYPE2TYPE:
                    kind = TYPE2TYPE[token.type]
                else:
                    kind = TYPE2TYPE.get((token.type, toplevel), Error)

                if token.type == JML.JML_KEYWORDS_TL_OPEN and toplevel:
                    waitForToplevelSemicolon = True

                yield (offset, kind, token.text)


TYPE2TYPE = {
    JML.AT_START:Comment.Preproc,
    JML.COMMENT_END:Comment.Preproc,
    JML.WS: Whitespace,
    JML.LPAREN: Punctuation,
    JML.RPAREN: Punctuation,
    JML.LBRACE: Punctuation,
    JML.RBRACE: Punctuation,
    JML.LBRACK: Punctuation,
    JML.RBRACK: Punctuation,
    JML.SEMI: Punctuation,
    JML.JAVA_MODIFIERS: Keyword,
    JML.JML_KEYWORDS_ALWAYS: Keyword,
    JML.JAVA_KEYWORDS: Keyword,
    (JML.JML_MODIFIERS, True): Keyword,
    (JML.JML_MODIFIERS, False): Name.Variable,
    (JML.JML_KEYWORDS_TL_EXPR, True): Keyword,
    (JML.JML_KEYWORDS_TL_EXPR, False): Keyword,
    (JML.JML_KEYWORDS_TL_OPEN, True): Keyword,
    (JML.JML_KEYWORDS_TL_OPEN, False): Name.Variable,
    (JML.JML_KEYWORDS_TL_NON_OPEN, True): Keyword,
    (JML.JML_KEYWORDS_TL_NON_OPEN, False): Name.Variable,
    JML.IDENTIFIER: Name.Variable,
    JML.NUM_LITERALS: Number,
    JML.COMMENT: Comment,
    JML.ASSIGN: Operator,
    JML.GT: Operator,
    JML.LT: Operator,
    JML.BANG: Operator,
    JML.TILDE: Operator,
    JML.QUESTION: Operator,
    JML.COLON: Operator,
    JML.EQUAL: Operator,
    JML.LE: Operator,
    JML.GE: Operator,
    JML.NOTEQUAL: Operator,
    JML.AND: Operator,
    JML.OR: Operator,
    JML.INC: Operator,
    JML.DEC: Operator,
    JML.ADD: Operator,
    JML.SUB: Operator,
    JML.MUL: Operator,
    JML.DIV: Operator,
    JML.BITAND: Operator,
    JML.BITOR: Operator,
    JML.CARET: Operator,
    JML.MOD: Operator,
    JML.ADD_ASSIGN: Operator,
    JML.SUB_ASSIGN: Operator,
    JML.MUL_ASSIGN: Operator,
    JML.DIV_ASSIGN: Operator,
    JML.AND_ASSIGN: Operator,
    JML.OR_ASSIGN: Operator,
    JML.XOR_ASSIGN: Operator,
    JML.MOD_ASSIGN: Operator,
    JML.LSHIFT_ASSIGN: Operator,
    JML.RSHIFT_ASSIGN: Operator,
    JML.URSHIFT_ASSIGN: Operator,
    JML.ARROW: Operator,
    JML.COLONCOLON: Operator,
    JML.DOTDOT: Operator,
    JML.EQUIVALENCE: Operator,
    JML.ANTIVALENCE: Operator,
    JML.IMPLIES: Operator,
    JML.IMPLIESBACKWARD: Operator,
    JML.LOCKSET_LEQ: Operator,
    JML. LOCKSET_LT: Operator,
    JML.ST: Operator,
    JML.AT: Operator,
    JML.ELLIPSIS: Operator,
    JML.ERROR_CHAR: Error,
}


class JmlConditionalChecker:
    def __init__(self, active_keys=None):
        # currently activated keys
        self.active_keys: List[str] = set(
            active_keys) if active_keys else list()
        self.cache: Dict[str, bool] = {}

    def is_active(self, text: str) -> bool:
        at_sign = text.find('@')
        keys = text.strip()[2: at_sign]
        cached_value = self.cache.get(keys, None)
        if cached_value:
            return cached_value

        conditions = keys.split("/(?=[+-])/")
        result = self.is_active_for_conditions(conditions)
        self.cache[keys] = result
        return result

    def is_active_for_conditions(self, conditions: List[str]) -> bool:
        # a JML annotation with at least one positive-key is only included
        plus_key_found = False
        # if at least one of these positive keys is enabled
        enabled_plus_key_found = False

        # a JML annotation with an enabled negative-key is ignored (even if there are enabled positive-keys).
        enabled_negative_key_found = False

        for marker in conditions:
            if marker is None or marker == "":
                continue
            is_positive = marker[0] == '+'
            is_negative = not is_positive
            k = marker[1:].lower()
            is_enabled = any((x == k for x in self.active_keys))
            plus_key_found = plus_key_found or is_positive
            enabled_plus_key_found = enabled_plus_key_found or is_positive and is_enabled
            enabled_negative_key_found = enabled_negative_key_found or is_negative and is_enabled
            if "-" == marker or "+" == marker:  # old deprecated conditions
                return False

        return (not plus_key_found or enabled_plus_key_found) and not enabled_negative_key_found

    def set_keys_from_token(self, token: Token):
        """Given a JML_SET_KEY token, this method sets the active keys for deciding the inclusion of
        conditional JML annotation comments. """
        # JML_SET_KEY: '//-*- jml-keys: ' (IDENTIFIER)* '-*-';
        prefix = "//-*- jml-keys: "
        suffix = ""  # "-*-"
        text = token.text
        if text:
            content = text.substring(len(prefix), text.length - len(suffix))
            keys = content.split("[ ,]")
            self.active_keys = [k.strip().lowercase() for k in keys]
            self.cache.clear()
